<?php

namespace Drupal\preprocessors\Service;

use Drupal\preprocessors\PluginManager\PreprocessorsPluginManager\PreprocessorsPluginManager;

/**
 * Provides an object to manage core functionality of the module.
 *
 * @package Drupal\preprocessors\Service
 */
final class PreprocessorsManager {

  /**
   * Preprocessor Plugin Manager injected through DI.
   *
   * @var \Drupal\preprocessors\PluginManager\PreprocessorsPluginManager\PreprocessorsPluginManager
   */
  private $preprocessorsPluginManager;

  /**
   * Array of plugin definitions to work with, stored statically.
   *
   * This variable is static to ensure that it is only stored once.
   *
   * @var array
   */
  private static $preprocessorDefinitions;

  /**
   * PreprocessorsManager constructor.
   *
   * @param \Drupal\preprocessors\PluginManager\PreprocessorsPluginManager\PreprocessorsPluginManager $preprocessorsPluginManager
   *   Preprocessor Plugin Manager injected through DI.
   */
  public function __construct(PreprocessorsPluginManager $preprocessorsPluginManager) {
    $this->preprocessorsPluginManager = $preprocessorsPluginManager;
  }

  /**
   * Preprocess variables using Preprocessor Plugins.
   *
   * @param array $variables
   *   Array of default variables obtained from the render instance.
   * @param string $hook
   *   The name of the theme hook being invoked. This is not always accurate.
   * @param array $info
   *   Array of information on the current render instance. This parameter is
   *   not even documented on Drupal's main site, but the module's functionality
   *   relies on it. If this breaks, everything breaks.
   */
  public function preprocess(array &$variables, string $hook, array $info): void {

    // Get the template name of the current render instance.
    $template = $info['template'] ?? NULL;

    // Do nothing if this info is not set.
    if ($template === NULL) {
      return;
    }

    // Set the preprocessor definitions as a static variable.
    if (static::$preprocessorDefinitions === NULL) {
      static::$preprocessorDefinitions = $this->preprocessorsPluginManager->getDefinitions();
    }

    // Get template parts.
    // Example: ['node', 'node--page'].
    $templateParts = $this->getTemplateParts($template);

    // With each part of the template, get Preprocessors matching the part.
    foreach ($templateParts as $template) {
      if (isset(static::$preprocessorDefinitions[$template])) {

        // Get the array of plugin definitions matching this template.
        $preprocessors = static::$preprocessorDefinitions[$template];

        // Load all of the plugins found that match, and apply them to data.
        foreach ($preprocessors as $plugin_id => $definition) {

          // Create plugin instance.
          $preprocessor = $this->preprocessorsPluginManager->createPreprocessorInstance($plugin_id, $template);

          // If created successfully, run preprocessing on variables.
          if ($preprocessor !== NULL) {
            $preprocessor->preprocess($variables, $hook, $info);
          }
        }
      }
    }
  }

  /**
   * Get the template parts segmented incrementally.
   *
   * In a template named "node--page", we want to make sure that Preprocessors
   * that target both the "node" template and the "node--page" template run.
   *
   * An array will be returned containing
   * each part of the template incrementally appended to each other.
   *
   * Example: A template like "node--page" will return an array
   * with the following values:
   * ['node', 'node--page'].
   *
   * The incremental  array form is used in many of the module's
   * features.
   *
   * @param string $template
   *   Template to get parts for. i.e. "node--page".
   *
   * @return mixed
   *   Return array of template parts.
   */
  private function getTemplateParts(string $template): array {

    // Explode into array of elements, using "." as a delimiter.
    // Example: "entity.node" becomes ['entity', 'node'].
    $parts = explode('--', $template);

    // Otherwise, we'll create the desired incremental format.
    // @see this function's documentation for more information.
    $result = [];

    foreach ($parts as $i => $part) {
      $needle = $part;

      if ($i !== 0) {
        $needle = '';
        for ($c = 0; $c <= $i; $c++) {
          $needle .= $parts[$c];
          if ($c !== $i) {
            $needle .= '--';
          }
        }
      }
      $result[] = $needle;
    }

    // Return the formatted result.
    return $result;
  }

}
